################################################################################
#                                                                              #
#                                                                              #
# Raster I/O, Leaf Area Index, Gap-Filling                   #
# ====================================                                         #
#                                                                              #
# Francesco Vuolo <francesco.vuolo@boku.ac.at>                                 #
# University of Natural Resources and Life Sciences, BOKU, Vienna                                                                              #
################################################################################
# Input data: Sentinel-2 data Leaf Area Index (LAI) and Cloud Masks (SCL)
# Output data: Gap-filled time series of LAI

# Remove all objects from Environment

rm(list = ls())

# =========
# Load packages:
# =========

# LOCAL FILE PATH MODIS PACKAGE path_to_MODIS<-paste0('MODIS_0.10-33.tar.gz')

if ( ! require(raster) )         { install.packages("raster");                                        library(raster)      }
if ( ! require(MODIS) )          { install.packages("MODIS");                                         library(MODIS)}
#if ( ! require(MODIS) )          { install.packages(path_to_MODIS, repos = NULL, type="source");      library(MODIS)       }


# ============================
# Set and create sub directories for input/output 
# ============================

indir  <- 'E:/ESA_LTC22_data/raw/'
outdir <- 'E:/ESA_LTC22_data/gap-filled/'

# Create directory to store gap-filled data 
dir.create(outdir,recursive=TRUE,showWarnings=FALSE)

# ============================
# Locate and Load Scene Classification raster for LAI cloud masking
# ============================
# Create RasterLayer object from a file 
filelist  <- list.files(path=indir,pattern=paste0("*.tif$") ,full.names=TRUE,recursive=TRUE) 

# filter for cloud mask files (SCL, sen2cor scene classification layer)
infls_SCL <- filelist[grep('SCL', filelist)]

# Mask for valid data range for LAI and clouds
for(i in seq_along(infls_SCL)){#i<-1
    
# read SCL
    SCL      <- raster(infls_SCL[i])
    
    SCL_val  <- getValues(SCL)
    
# find and read corresponding LAI    
    LAI <- raster(sub('SCL', 'LAI', infls_SCL[i])) 
    
    var_val  <- getValues(LAI) #  returns all values for the Raster
    
#    hist(var_val,breaks= 100, xlim = c(-500,3000))
    
# Management of extreme cases 
    
    # tolerance = 200 (0.2)
    var_val[var_val >  -200   & var_val < 0] = 0
    var_val[var_val >  8000   & var_val <= 8200] = 8000
    var_val[var_val < -200 | var_val > 8200] = NA
    
# hist(var_val,breaks= 100, xlim = c(-500,3000))
    
# SCL Label	Classification
    # 0	NO_DATA
    # 1	SATURATED_OR_DEFECTIVE
    # 2	CAST_SHADOWS
    # 3	CLOUD_SHADOWS
    # 4	VEGETATION
    # 5	NOT_VEGETATED
    # 6	WATER
    # 7	UNCLASSIFIED
    # 8	CLOUD_MEDIUM_PROBABILITY
    # 9	CLOUD_HIGH_PROBABILITY
    # 10	THIN_CIRRUS
    # 11	SNOW or ICE 
    
    var_val[SCL_val == 0 | SCL_val == 3 | SCL_val == 8 | SCL_val == 9 | SCL_val == 10 ] = NA
    
    filenameOut <- sub('SCL', 'LAI_masked', infls_SCL[i])

    writeRaster(x= setValues(LAI,var_val),
                filename = filenameOut,
                NAflag = 20000,
                format = "GTiff",
                datatype = 'INT2U',
                overwrite = TRUE,
                options = "COMPRESS=DEFLATE")
    
    rm(SCL,SCL_val,LAI,var_val)
    
    #file.remove(name)
    
  }

# ============================
# Gap filling and smoothing
# ============================
# Locate the files to gap - fill 

filelist <- list.files(path=indir,pattern=paste0("*.tif$") ,full.names=TRUE,recursive=TRUE) 
infls_LAI <-filelist[grep('masked', filelist)]

#Organize (MODIS) Files in Preparation for Stacking
infls_LAI<-preStack(files=infls_LAI)

# doy<-format(as.Date('20220830',format = "%Y%m%d"), format = "%j") 

# Handle Input and Output Dates Used for Filtering
# With this function you define: 
# - the period to be filtered (begin:end) BEGIN: Default is from earliest input dataset; END: Default to the end of the input dataset.
# - the output temporal resolution (nDays)
    
  timeInfo=orgTime(
        basename(infls_LAI),
        nDays = 10,
        begin=NULL, # expressed with year and DOY
        end= NULL,  # expressed with year and DOY
        pillow=2,
        pos1=26,# position in the filename indicating the date of acquisition - start (pos1) and end (pos2)
        pos2=33,
        format = "%Y%m%d"
      )

      show(timeInfo$inputLayerDates)
      show(timeInfo$outputLayerDates)


# Filter Vegetation Index with Modified Whittaker Approach
info <- whittaker.raster(infls_LAI, 
                timeInfo = timeInfo,
                lambda = 5000,
                nIter = 1,
                outputAs = "single",
                collapse = FALSE,
                prefixSuffix = c("LAI_i1"),
                outDirPath = outdir,
                outlierThreshold = NULL,
                mergeDoyFun = "max",
                overwrite=TRUE)

# ============================
# plot results  
# ============================
# Locate the files to gap - fill 

#stack raw data
filelist  <- list.files(path=indir,pattern=paste0("*LAI.tif$") ,full.names=TRUE,recursive=TRUE) 
raw<-stack(filelist)

# stack gap-filled data
filelist <- list.files(path=outdir,pattern=paste0("*.tif$") ,full.names=TRUE,recursive=TRUE) 
filelist <-filelist[grep('LAI_i1', filelist)]
gp<-stack(filelist)

x <- 1 #numeric vector representing cell numbers

raw_y <- as.numeric(extract(raw, x)) # extract raw data
raw_x <- timeInfo$inDoys

gp_y<-as.numeric(extract(gp, x)) # extract gapfilled data
gp_x<-as.numeric(format(timeInfo$outputLayerDates, format = "%j"))

# plot
plot(raw_x,raw_y/1000,xlab = "DOY", ylab = "LAI", pch = 1, col=8,
     font.main = 1, # Title font style
     font.sub  = 2, # Subtitle font style
     font.axis = 4, # Axis tick labels font style
     font.lab  = 4) # Font style of X and Y axis labels
lines(gp_x,gp_y/1000, col=1, lwd =2)




#legend("topright", legend = c("l=50", "l=500", "l=1000", "l=3000"),
       lwd = 2, col = c("black", "red","green","blue"))


# filelist <- list.files(path=outdir,pattern=paste0("*.tif$") ,full.names=TRUE,recursive=TRUE) 
# filelist <-filelist[grep('LAI_iter3', filelist)]
# gp_1<-stack(filelist)
# 
# gp_1_y<-as.numeric(extract(gp_1, x)) # extract gapfilled data
# gp_1_x<-as.numeric(format(timeInfo$outputLayerDates, format = "%j"))
# 
# # plot
# plot(raw_x,raw_y)
# lines(gp_x,gp_y)
# lines(gp_1_x,gp_1_y)